#' @name lda
#' @title Linear Discriminant Analysis.

#' @description For a given training dataset of observations containing :
#' \itemize{
#'  \item{\strong{explanatory variables} (\emph{quantitative or qualitative})}
#'  \item{and a \strong{classification variable} defining groups of observations}
#'  }
#'  The lda() function performs :
#' \enumerate{
#'   \item{ \strong{Feature selection} \emph{OPTIONAL}:\cr
#'   selects a subset of the quantitative variables to discriminate the classes\cr
#'   \emph{Inspired from SAS PROC STEPDISC (see ?stepdisc)}}
#'
#'   \item{\strong{Linear Discriminant Analysis} :\cr
#'   develops a discriminant criterion to classify each observation into one of the groups.\cr
#'   \emph{Inspired from SAS PROC DISCRIM}}
#'  }
#'  The \code{model} obtained can be applied on a test dataset to predict the class of each observation. (\emph{see ?predict})\cr
#' @details \itemize{
#'   \item Categorical variables will be automatically converted into dummy variables.
#' }
#' @param data \emph{data.frame} \cr the training dataset containing explanatory and classification variables.
#' @param target \emph{character} \cr the classification variable name.
#' @param stepdisc \emph{boolean} \cr Does feature selection has to be performed ?\cr
#' FALSE \cr
#' TRUE (\emph{DEFAULT})
#' @param method \emph{character} \cr If stepdisc==TRUE, the feature selection method to be used :
#' \describe{
#'  \item{'forward'}{\emph{alternative :} 'F' | 'f' | 'forw'}
#'  \item{'backward'}{\emph{alternative :} 'B' | 'b' | 'back'}
#'  \item{'stepwise'}{\emph{alternative :} 'S' | 's' | 'step'}
#' }
#' @param verbose \emph{boolean} \cr TRUE : print intermediary steps during feature selection (\emph{DEFAULT})
#' \cr FALSE : print only feature selection' summary.
#' @param pval_cut \emph{numeric} \cr Specifies the significance level for adding or retaining variables in stepwise variable selection (\emph{DEFAULT = 0.05})
#'
#' @usage
#' \strong{WITHOUT feature selection}
#' lda(data, target)
#'
#' \strong{WITH stepwise feature selection} WITH intermediary outputs
#' lda(data, target, stepdisc=TRUE, method)
#'
#' \strong{WITH stepwise feature selection} WITHOUT intermediary outputs
#' lda(data, target, stepdisc=TRUE, method, verbose=FALSE)
#'
#' @return Returns a \code{model} object containing :
#' \itemize{
#'  \item \code{features} (if stepdisc = TRUE)
#'  \item \code{levels}
#'  \item \code{const}
#'  \item \code{coefs}
#'  \item \code{Pooled Covariance} \emph{list}
#'    \itemize{
#'      \item Information \emph{data.frame}
#'      \item Matrix \emph{matrix}
#'    }
#'  \item \code{summary} \emph{list}
#'    \itemize{
#'      \item Dataset Information \emph{data.frame}\describe{
#'        \item{Total Sample Size}{}
#'        \item{Variables}{}
#'        \item{Classes}{}
#'        }
#'      \item Class Level Information \emph{data.frame}\describe{
#'        \item{frequency}{}
#'        \item{proportion}{}
#'        }
#'    }
#' }
#' @examples
#'
#' data(HRdata) # Load data
#'
#' # WITHOUT feature selection
#' lda(data = HRdata, target = 'left')
#'
#' # WITH feature selection - STEPWISE approach and WITH intermediary outputs
#' lda(data = HRdata, target = 'left', stepdisc = TRUE, method = 's')
#'
#' # WITH feature selection - BACKWARD approach and WITHOUT intermediary outputs
#' lda(data = HRdata, target = 'left', stepdisc = TRUE, method = 'backward', verbose=FALSE)
#' @source
#' \url{http://www.math.wpi.edu/saspdf/stat/chap25.pdf}
#' \url{https://support.sas.com/documentation/cdl/en/statugdiscrim/61779/PDF/default/statugdiscrim.pdf}
#' \url{https://eric.univ-lyon2.fr/~ricco/cours/slides/analyse_discriminante.pdf}
#' @export


#lda <- function(data, target, stepdisc = FALSE, method="stepwise", verbose=TRUE){

#add pval_cut to the parameters
lda <- function(data, target, stepdisc = FALSE, method="stepwise", verbose=TRUE, pval_cut=0.05){

  # [FUNCTION] Compute the covariance matrix
  cov.mat <- function(x){
    (nrow(x)-1)*cov(x)
  }

  # [FUNCTION] Compute the intercept
  calc.const <- function (x){
    -0.5*t(colMeans(x))%*%solve(pooled.cov.mat)%*%colMeans(x)+log(nrow(x)/length(train_Y))
  }

  # [FUNCTION] Compute weight for each features
  calc.coefs <- function(x){
    solve(pooled.cov.mat)%*%colMeans(x)
  }

  #-------------------------------------
  # --------------- MAIN ---------------
  #-------------------------------------

  model <- list()

  # [DATA] Split features and target
  train_X <- data[,colnames(data) != target]
  train_Y <- data[, target]
  if(is.factor(train_Y) == FALSE){
    train_Y <- as.factor(train_Y)
  }

  # [CONTROL] Check if there are some categorical features
  # One hot encoding : generate dummy features for each modality
  if(sum(sapply(train_X, is.factor)) != 0){
    to_be_encoded <- sapply(data.frame(train_X), is.factor)
    train_X <- data.frame(model.matrix( ~ ., data=train_X,
                                        contrasts.arg = lapply(data.frame(train_X[,to_be_encoded]),
                                                               contrasts,
                                                               contrasts = TRUE
                                                               )
                                        )
                          )
    train_X$X.Intercept. <- NULL # Delete the intercept column
    if(verbose) cat('\nCategorical features have been encoded into binary variables.\n')
  }

  # [CONTROL] Check the method name.
  # Stop if unknown
  if (method %in% c('F','f','forw','forward')){ method  <-  "forward" }
  else if (method %in% c('S','s','step','stepwise')){ method  <-  "stepwise" }
  else if (method %in% c('B','b','back','backward')){ method  <-  "backward" }
  else{stop(cat("Unknown method.\nPlease choose among{'forward','backward','stepwise'}."))}

  # [CONTROL] Check parameter stepdisc and launch feature selection
  if(stepdisc==TRUE){
    #selected_features <- stepdisc(var=train_X, class_vec=train_Y, pval_cut=0.05, method=method, verbose=verbose)
    selected_features <- stepdisc(var=train_X, class_vec=train_Y, pval_cut=pval_cut, method=method, verbose=verbose)
    # Update model
    train_X <- train_X[,selected_features]
    model$features <- selected_features
    # Print
    cat(paste("\n",length(selected_features),"features selected with",method,"approach :\n\n"))
    print(selected_features)
  }

  model$summary <- list()
  variables <- ncol(train_X)
  classes <- nlevels(train_Y)
  sample_size <- nrow(train_X)
  frequency <- tapply(train_Y, train_Y, length)
  proportion <- frequency/length(train_Y)

  model$summary$"Dataset Information" <- data.frame(rbind(sample_size, variables, classes))
  rownames(model$summary$"Dataset Information") <- c("Total Sample Size", "Variables", "Classes")
  colnames(model$summary$"Dataset Information") <- c("Count")

  model$summary$"Class Level Information" <- data.frame(frequency, proportion)


  # Compute covariance matrix for each class
  liste.cov.mat <- by(train_X, train_Y, cov.mat)


  #Pooled covariance matrix
  pooled.cov.mat <- (1.0 / (length(train_Y) - nlevels(train_Y))) * Reduce('+', liste.cov.mat)
  det <- determinant(pooled.cov.mat, logarithm = TRUE)$modulus
  rank <- as.integer(ncol(pooled.cov.mat))

  model$"Pooled Covariance"$"Information" <- data.frame(rbind(rank, det),row.names = c("Covariance Matrix Rank","Natural Log of the Determinant of the Covariance Matrix"))
  colnames(model$"Pooled Covariance"$"Information") <- "Value"
  model$"Pooled Covariance"$"Matrix" <- pooled.cov.mat


  # Add to the object  the model parameters (used by predict function)
  model$const <- by(train_X, train_Y, calc.const)
  model$coefs <- by(train_X, train_Y, calc.coefs)
  model$levels <- levels(train_Y)

  #class(model) <- "Linear Discriminant Analysis model"
  #simplifying the class name
  class(model) <- "LDA"
  return(model)
}


