#' @title Classes prediction
#' @description The predict() method use the \code{model} obtained from the lda method \strong{to classify each observations of a test dataset.}
#' @param model \emph{object of class "Discriminant analysis model"} \cr the \code{model} obtained from the lda method
#' @param test_X \emph{data.frame} \cr the explanatory variables of the testing dataset.
#' @details \itemize{
#'   \item Categorical variables will be automatically converted into dummy variables.
#'   \item If feature selection has been used to fit the model, the same variables will be used.
#' }
#' @return Returns a vector of predicted class for each observation of \code{test_X.}
#' @examples
#' predicted_label <- predict(model = model, test_X = test_X)
#' @export

#predict <- function(model, test_X)
#associate predict to the LDA class
predict.LDA <- function(model, test_X){
  coefs <- model$coefs
  const <- model$const

  # [FUNCTION] Compute class score
  score <- function(test_X){
    sapply(coefs, function(x) as.numeric(test_X)%*%x)+as.numeric(const)
  }

  # [CONTROL] Check if there are some categorical features
  # One hot encoding : generate dummy features for each modality
  if(sum(sapply(test_X, is.factor)) != 0){
    to_be_encoded <- sapply(data.frame(test_X), is.factor)
    test_X <- data.frame(model.matrix( ~ ., data=test_X,
                                       contrasts.arg = lapply(data.frame(test_X[,to_be_encoded]),
                                                              contrasts,
                                                              contrasts = TRUE
                                                              )
                                       )
                         )
    test_X$X.Intercept. <- NULL # Delete the intercept col
    cat('\nCategorical features have been encoded into binary variables.\n')
  }

  # [CONTROL] Check in the model if feature selection has been done
  if(exists("features", where = model)){
    cat('\nThis model has been obtained with feature selection.\n Same variables will be used to predict.')
    test_X <- test_X[,model$features]
  }


  # classes allocation according to the given scores
  scores <- apply(test_X, 1, score)
  classes <- apply(scores, 2, which.max)
  predictions <- model$levels[classes]
  class(predictions) <- 'pred'
  return(predictions)
}
