#' @title Feature selection.
#' @description
#' Given a classification variable and several quantitative variables, the stepdisc function
#' selects a subset of the quantitative variables to discriminate the classes.\cr\cr
#' 3 approaches are available :
#' \describe{
#'  \item{\strong{Forward selection}}{
#'    \enumerate{
#'      \item Begins with \strong{no variables} in the model.
#'      \item At each step, the variable that contributes most to the discriminatory power of the model (measured by Wilks’
#' lambda) is entered.
#'      \item When none of the unselected variables meet the entry criterion, the forward selection process stops.
#'     }}
#'  \item{\strong{Backward elimination}}{
#'    \enumerate{
#'      \item Begins with \strong{all variables} in the model.
#'      \item At each step, the variable that contributes least to the
#'      discriminatory power of the model (measured by Wilks’ lambda)  is removed.
#'      \item When all remaining variables meet the criterion to stay in the model, the backward elimination process stops
#'     }}
#'  \item{\strong{Stepwise selection}}{
#'    \enumerate{
#'      \item Begins with \strong{no variables} in the model
#'      \item At each step, the model is examined.
#'        \itemize{
#'          \item If the variable in the model that contributes least to the discriminatory power of the model (measured by Wilks’ lambda) fails to meet the criterion to stay, then that variable is removed.
#'          \item Otherwise, the variable not in the model that contributes most to the discriminatory power of the model is entered.
#'        }
#'      \item When all variables in the model meet the criterion to stay and none of the
#'      other variables meet the criterion to enter, the stepwise selection process stops.
#'     }}}\cr
#'  \emph{Inspired from SAS PROC STEPDISC}
#' @param var \emph{data.frame}
#' @param class_vec \emph{character}
#' @param method \emph{character}
#' \describe{
#'  \item{'forward'}{\emph{alternative :} 'F' | 'f' | 'forw'}
#'  \item{'backward'}{\emph{alternative :} 'B' | 'b' | 'back'}
#'  \item{'stepwise'}{\emph{alternative :} 'S' | 's' | 'step'}
#' }
#' @param pval_cut \emph{character}
#' @param verbose \emph{boolean} \cr TRUE : print intermediary steps during feature selection (\emph{DEFAULT})
#' \cr FALSE : print only feature selection' summary.
#' @return Returns the list of \code{features} selected.
#' @examples
#'
#' data(HRdata) # Load data
#' target ='left' # define target
#'
#' # Split data
#' train_X <- HRdata[,colnames(data) != target]
#' train_Y <- HRdata[, target]
#'
#' # Feature selection using stepwise approach
#' stepdisc(var = train_X, class_vec = train_Y, method = 'stepwise', pval_cut=0.05)
#'
#' #' # Feature selection using forward approach
#' stepdisc(var = train_X, class_vec = train_Y, method = 'f', pval_cut=0.05)

#' @source
#' \url{https://support.sas.com/documentation/cdl/en/statug/63347/HTML/default/viewer.htm#stepdisc_toc.htm}
#' \url{https://support.sas.com/documentation/cdl/en/statugstepdisc/61834/PDF/default/statugstepdisc.pdf}
#' \url{https://eric.univ-lyon2.fr/~ricco/cours/slides/filtrage_predicteurs.pdf}
#' @export


stepdisc <- function(var, class_vec, method, pval_cut=0.05, verbose=TRUE){

  cov_k  <-  function(x){
    data_k <- as.matrix(x)
    n_k <- nrow(data_k)
    cov_k <- cov(data_k)
    #modif., estimation (n_k - 1) after cov()
    #cov_k <- n_k*cov_k
    cov_k <- (n_k - 1) * cov_k
  }

  intraclass <- function(dataset, classes){
    covlist <- by(dataset, classes, cov_k)
    return(Reduce("+",covlist)/nrow(dataset))
  }

  lambda_wilks <- function(dataset,totalcov,intracov){
    return(det(as.matrix(intracov[colnames(dataset),colnames(dataset)]))/det(as.matrix(totalcov[colnames(dataset),colnames(dataset)])))
  }

  ftest <- function(lw1,lw2,j,n,K){
    # compute partial R2, F-value and P-value at step j (ie : j variables in the model)
    # lw1 is the wilks lambda at step j-1, lw2 is the wilks lambda at step j
    # n number of individuals, K number of classes
    r2 <- (lw1-lw2)/lw1
    f <- (lw1/lw2-1)*(n-K-j+1)/(K-1)
    p <- 1-pf(f,K-1,n-K-j+1)
    return(list(R2=r2,Fval=f,pval=p))
  }

  step_asc <- function(j, lw_prev, varin, varout, class_vec, pval_cut, summary, totalcov, intracov, method){
    # forward step : for a given model, check if adding a new variable brings a significant better discrimination.
    # 1) Take the most discriminating variable from variable set (highest F-value)
    # 2) Check if it brings significant discrimination (p-value < cut)

    if(verbose) cat(paste("\n ************ Step ",j," , forward ************  \n\n"))

    lw_vec <- double(0) #contains F-values
    R2_vec <- double(0) #contains p-values
    F_vec <- double(0) #contains F-values
    Pval_vec <- double(0) #contains p-values

    n <- nrow(varout)
    K <- nlevels(class_vec)

    for(i in 1:ncol(varout)){
      modeltest <- cbind(varin,varout[,i])
      colnames(modeltest)[j] <- colnames(varout)[i]
      #print(colnames(modeltest))
      lw <- lambda_wilks(modeltest,totalcov, intracov)
      #print(lw)
      res_ftest <- ftest(lw_prev,lw,j,n,K) # lw_prev is wilks lambda at step j-1
      r2 <- res_ftest$R2
      f <- res_ftest$Fval
      p <- res_ftest$pval
      lw_vec <- c(lw_vec,lw)
      R2_vec <- c(R2_vec,r2)
      F_vec <- c(F_vec,f)
      Pval_vec <- c(Pval_vec,p)
    }

    step_summary <- matrix(0,nrow=ncol(varout),ncol=3)
    rownames(step_summary) <- colnames(varout)
    colnames(step_summary) <- c("partial R2", "F-value", "p-value")
    step_summary[,1] <- R2_vec
    step_summary[,2] <- F_vec
    step_summary[,3] <- Pval_vec

    if(verbose) print(step_summary)

    if(Pval_vec[which.max(F_vec)]<pval_cut){
      if(verbose) cat(paste("\n Variable ",colnames(varout)[which.max(F_vec)]," will be entered \n"))
      summary <- rbind(summary,c(j,colnames(varout)[which.max(F_vec)],"",R2_vec[which.max(F_vec)],max(F_vec),Pval_vec[which.max(F_vec)],lw_vec[which.max(F_vec)]))
      varin <- cbind(varin,varout[,which.max(F_vec)]) # add variable to model
      colnames(varin)[j] <- colnames(varout)[which.max(F_vec)]
      varout.names <- colnames(varout)[-which.max(F_vec)]
      varout <- as.matrix(varout[,-which.max(F_vec)]) # remove variable from variable set
      colnames(varout) <- varout.names

      if(verbose) cat("\n Variable(s) that have been entered : \n")
      if(verbose) print(colnames(varin))

      # If method == "stepwise": launch backward at step j
      if(method == "stepwise") {
        return(step_des(j,lw_vec[which.max(F_vec)],varin,varout,class_vec,pval_cut,summary,totalcov,intracov,method))
      }
      # If method == "ascendante": launch ascendante at step j+1
      if(method == "forward"){
        if(j<nrow(totalcov)){
          step_asc(j+1,lw_vec[which.max(F_vec)],varin,varout,class_vec,pval_cut,summary,totalcov,intracov,method)
        }
        else{
          if(verbose) cat("\n\n No further steps are possible \n\n")
          if(nrow(summary)>0){
            if(verbose) print(summary)
            return(colnames(varin))
          }
        }
      }
    }  else{
      if(verbose) cat("\n\n No more variables can be entered \n\n")
      if(nrow(summary)>0){
        if(verbose) print(summary)
        return(colnames(varin))
      }
    }
  }

  step_des <- function(j, lw_prev, varin, varout, class_vec, pval_cut, summary, totalcov, intracov, method){
    # backward step : for a given model, check if removing a variable brings a significant loss of discrimination.
    # 1) Take the variable whose removal brings the lowest loss of discrimination (lowest F-value)
    # 2) Check if the loss of discrimination is significant (p-value > cut)

    if(verbose) cat(paste("\n ************ Step ",j," , backward ************  \n\n"))

    lw_vec <- double(0) #contains wilks lambda values
    R2_vec <- double(0) #contains partial R2 values
    F_vec <- double(0) #contains F-values
    Pval_vec <- double(0) #contains p-values

    n <- nrow(varin)
    K <- nlevels(class_vec)

    for(i in 1:ncol(varin)){
      modeltest <- as.matrix(varin[,-i])  #remove a variable from model
      if(j>1){
        lw <- lambda_wilks(modeltest,totalcov,intracov)
      }
      else{
        lw <- 1
      }
      res_ftest <- ftest(lw,lw_prev,j,n,K)
      r2 <- res_ftest$R2
      f <- res_ftest$Fval
      p <- res_ftest$pval
      lw_vec <- c(lw_vec,lw)
      R2_vec <- c(R2_vec,r2)
      F_vec <- c(F_vec,f)
      Pval_vec <- c(Pval_vec,p)
    }

    step_summary <- matrix(0,nrow=ncol(varin),ncol=3)
    rownames(step_summary) <- colnames(varin)
    colnames(step_summary) <- c("partial R2", "F-value", "p-value")
    step_summary[,1] <- R2_vec
    step_summary[,2] <- F_vec
    step_summary[,3] <- Pval_vec

    if(verbose) print(step_summary)

    if(Pval_vec[which.min(F_vec)]>pval_cut){
      if(verbose) cat(paste("\n Variable ",colnames(varin)[which.min(F_vec)]," will be removed \n "))
      summary <- rbind(summary,c(j,"",colnames(varin)[which.min(F_vec)],R2_vec[which.min(F_vec)],min(F_vec),Pval_vec[which.min(F_vec)],lw_vec[which.min(F_vec)]))

      #modif. which.min is better here for varin
      varout <- cbind(varout,varin[,which.min(F_vec)]) # put back variable to variable set

      varin.names <- colnames(varin)[-which.min(F_vec)]
      varin <- as.matrix(varin[,-which.min(F_vec)]) # remove variable from model

      #modif.
      colnames(varin) <- varin.names

      #modif. - which.min is not valid here for varin because it is removed
      #print(colnames(varin))
      #print(which.min(F_vec))

      #modif. - moved above
      #varout <- cbind(varout,varin[,which.min(F_vec)]) # put back variable to variable set

      if(method == "stepwise"){
        return(step_asc(j-1,lw_vec[which.min(F_vec)],varin,varout,class_vec,pval_cut,summary,totalcov,intracov,method))
      }
      if(method == "backward"){
        step_des(j-1,lw_vec[which.min(F_vec)],varin,varout,class_vec,pval_cut,summary,totalcov,intracov,method)
      }
    }
    else{
      if(verbose) cat("\n No variable can be removed \n ")
      if(ncol(varout)>0 & method == "stepwise") {
        return(step_asc(j+1,lw_prev,varin,varout,class_vec,pval_cut,summary,totalcov,intracov,method))
      }
      else{
        if(verbose) cat("\n\n No further steps are possible. \n\n")
        if(nrow(summary)>0){
          if(verbose) print(summary)
          return(colnames(varin))

        }
      }
    }
  }

  #-------------------------------------
  # --------------- MAIN ---------------
  #-------------------------------------

  summary <- matrix(0,nrow=0,ncol=7) # to keep track of the informations during the process
  colnames(summary) <- c("Step","Entered","Removed","Partial R2","F-value","P-value","Wilk's lambda")

  #modif. -- check the degree of freedom
  #totalcov <- cov(var)
  #intracov <- intraclass(var,class_vec)

  #modif. --  biased estimator is right for the Wilks' lambda
  rr_n <- nrow(var)
  totalcov <- (rr_n - 1)/rr_n * cov(var)
  #not modified
  intracov <- intraclass(var,class_vec)


  # [CONTROL] Check the method name. (twice because this function could be autonomous)
  # Stop if unknown
  if (method %in% c('F','f','forw','forward')){ method  <-  "forward" }
  else if (method %in% c('S','s','step','stepwise')){ method  <-  "stepwise" }
  else if (method %in% c('B','b','back','backward')){ method  <-  "backward" }
  else{stop(cat("Unknown method.\nPlease choose among{'forward','backward','stepwise'}."))}

  if (method == "forward" | method == "stepwise"){
    j <- 1 #number of variables in the model at first step
    lw0 <- 1 #initial wilks lambda
    varin <- matrix(0,nrow(var),0) #variables inside the model
    varout <- var #variables outside the model
    features <- step_asc(j, lw0, varin, varout, class_vec, pval_cut, summary, totalcov, intracov, method)
  }
  else if(method == "backward"){
    j <- ncol(var) #number of variables in the model at first step
    lw0 <- lambda_wilks(var,totalcov,intracov)  #initial wilks lambda
    varin <- var #variables inside the model
    varout <- matrix(0,nrow(var),0) #variables outside the model
    features <- step_des(j, lw0, varin, varout, class_vec, pval_cut, summary, totalcov, intracov, method)
  }
  class(features) <- "Selected Features"
  return(features)
}
